// SystemService.cs
//
// GNOME Do is the legal property of its developers. Please refer to the
// COPYRIGHT file distributed with this source distribution.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

using System;
using System.Reflection;
using System.Diagnostics;
using System.Windows.Forms;

using Do.Platform;
using Do.Platform.ServiceStack;

using Microsoft.Win32;

namespace Do.Platform.Windows
{

	public class SystemService : AbstractSystemService, IInitializedService
	{
		const string AutoStartKey = "GNOMEDo";
		const string AutoStartKeyLocation = @"Software\Microsoft\Windows\CurrentVersion\Run";

		public void Initialize ()
		{
			SystemEvents.PowerModeChanged += PowerOnBatteryChanged;
		}

		void PowerOnBatteryChanged (object sender, PowerModeChangedEventArgs args)
		{
			if (args.Mode == PowerModes.StatusChange) {
				OnOnBatteryChanged ();
			}
		}

		public override bool GetOnBatteryPower ()
		{
			try {
				return SystemInformation.PowerStatus.PowerLineStatus == PowerLineStatus.Offline;
			} catch (Exception e) {
				Log<SystemService>.Error ("Could not GetOnBattery: {0}", e.Message);
				Log<SystemService>.Debug (e.StackTrace);
			}
			return false;
		}

		public override void EnsureSingleApplicationInstance ()
		{
			try {
				if (Process.GetProcessesByName (Process.GetCurrentProcess ().ProcessName).Length > 1) {
					Log<SystemService>.Debug ("No other application instance detected. Continue startup.");
				} else {
					Log<SystemService>.Debug ("Existing application instance detected. Summon and bail.");
					Summon ();
					System.Environment.Exit (0);
				}
			} catch (Exception e) {
				Log<SystemService>.Error ("Could not EnsureSingleApplicationInstance: {0}", e.Message);
				Log<SystemService>.Debug (e.StackTrace);
			}
		}

		public void Summon ()
		{
			Gdk.Threads.Enter ();
			Services.Windowing.SummonMainWindow ();
			Gdk.Threads.Leave ();
		}

		public override bool IsAutoStartEnabled ()
		{
			RegistryKey key = Registry.CurrentUser.OpenSubKey (AutoStartKeyLocation);

			if (key == null)
				return false;

			string value = ( string ) key.GetValue (AutoStartKey);
			if (value == null)
				return false;

			return (value == Assembly.GetExecutingAssembly ().Location);
		}

		public override void SetAutoStartEnabled (bool enabled)
		{
			RegistryKey key = Registry.CurrentUser.CreateSubKey (AutoStartKeyLocation);

			try {
				if (IsAutoStartEnabled () == enabled)
					return;
				if (enabled)
					key.SetValue (AutoStartKey, Assembly.GetExecutingAssembly ().Location);
				else
					key.DeleteValue (AutoStartKey);
			} catch (Exception e) {
				Log<SystemService>.Error ("Failed to update autostart registry key: {0}", e.Message);
			}
		}
	}
}
