// TrayIconService.cs
//
// GNOME Do is the legal property of its developers. Please refer to the
// COPYRIGHT file distributed with this source distribution.
//  
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

using System;
using System.Drawing;
using System.Windows.Forms;

using Do.Platform;
using Do.Platform.ServiceStack;

namespace Do.Platform.Windows
{
	public class TrayIconService : IInitializedService
	{
		const int IconSize = 24;
		const int NotifyDelay = 2000;
		const string IconName = "gnome-do";

        /// <summary>
        /// The content menu for the system tray icon
        /// </summary>
        private static ContextMenu m_menu;

        /// <summary>
        /// The system tray icon itself
        /// </summary>
        private static NotifyIcon m_notifyicon;

		// Must remain static until IServices have their own preferences.
		static TrayIconPreferences Preferences { get; set; }

		// Need until IServices have their own preferences.
		public static bool Visible {
			get { return Preferences.IconVisible; }
			set { Preferences.IconVisible = value; }
		}

		NotificationHelper notifier;

		public TrayIconService ()
		{
            if (m_menu == null) {
                // TODO make the context menu
                m_menu = new ContextMenu ();
                m_menu.MenuItems.Add (0, new MenuItem ("this does nothing"));
            }

            if (m_notifyicon == null) {
                m_notifyicon = new NotifyIcon ();
                m_notifyicon.Text = "WINDOWS Do";
                m_notifyicon.Visible = false;
                m_notifyicon.Icon = new Icon (Properties.Resources.app_icon, 16, 16);
                m_notifyicon.ContextMenu = m_menu;
                m_notifyicon.BalloonTipIcon = ToolTipIcon.Warning;
            }

            notifier = new NotificationHelper (m_notifyicon);
			notifier.NotificationClosed += OnNotificationClosed;
		}

		public void Initialize ()
		{
			Preferences = new TrayIconPreferences ();
			Preferences.IconVisibleChanged += OnIconVisibleChanged;
			if (!Preferences.IconVisible) 
				Hide ();

			// Listen for notifications so we can show a notification bubble.
			Services.Notifications.Notified += OnNotified;
		}

		void OnIconVisibleChanged (object sender, EventArgs e)
		{
			if (Preferences.IconVisible)
				Show ();
			else
				Hide ();
		}

		void OnNotified (object sender, NotificationEventArgs e)
		{
			Services.Application.RunOnMainThread (() => {
				ShowNotification (e.Notification);
			});
		}

		void OnNotificationClosed (object sender, NotificationEventArgs e)
		{
			Services.Application.RunOnMainThread (() => {
				if (!Preferences.IconVisible) Hide ();
			});
		}

		// Only run on the main thread.
		void ShowNotification (Notification note)
		{
            Show ();

		    Services.Application.RunOnMainThread (() => {
			    notifier.Notify (note);
		    });
		}

		void Show ()
		{
            m_notifyicon.Visible = true;
		}

		void Hide ()
		{
            m_notifyicon.Visible = false;
		}
	}
}
