// ApplicationItem.cs
//
// GNOME Do is the legal property of its developers. Please refer to the
// COPYRIGHT file distributed with this source distribution.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

using System;
using System.IO;
using System.Linq;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text.RegularExpressions;

using Do.Universe;
using Do.Platform;

using IWshRuntimeLibrary;
using System.Diagnostics;

namespace Do.Universe.Windows {

	internal class ApplicationItem : Item, IApplicationItem {
		private static WshShell shell = new WshShell ();

		const string DefaultApplicationIcon = "applications-other";

		static IDictionary<string, ApplicationItem> Instances { get; set; }

		static ApplicationItem ()
		{
			Instances = new Dictionary<string, ApplicationItem> ();
		}

		public static ApplicationItem MaybeCreateFromDesktopItem (string path)
		{
			ApplicationItem appItem;

			if (path == null) throw new ArgumentNullException ("path");

			if (Instances.ContainsKey (path)) {
				appItem = Instances [path];
			} else {
				IWshShortcut item = null;
				try {
					item = (IWshShortcut) shell.CreateShortcut (path);
					appItem = new ApplicationItem (item);
				} catch (Exception e) {
					appItem = null;
					item = null;
					Log.Error ("Could not load desktop item: {0}", e.Message);
					Log.Debug (e.StackTrace);
				}

				if (appItem != null)
					Instances [path] = appItem;
			}
			return appItem;
		}

		public static ApplicationItem MaybeCreateFromCmd (string cmd)
		{
			if (string.IsNullOrEmpty (cmd))
				return null;

			List<ApplicationItem> appItems = new List<ApplicationItem> ();

			cmd = Regex.Escape (cmd);
			Regex regex = new Regex (string.Format ("(^| ){0}( |)", cmd));
			foreach (ApplicationItem item in Instances.Values) {
				string path = item.Location;
				try {
					if (path.StartsWith ("file://"))
						path = path.Substring ("file://".Length);

					path = Path.GetFileName (path);
				} catch {
					continue;
				}

				try {
					if (!string.IsNullOrEmpty (path) && !string.IsNullOrEmpty (item.Exec) &&
						(regex.IsMatch (path) || regex.IsMatch (item.Exec))) {
						appItems.Add (item);
					}
				} catch {
					// it failed, probably a null somewhere, we dont care really
				}
			}

			return appItems [0];
		}

		protected IWshShortcut item;
		string name, description, icon, mimetype;
		IEnumerable<string> categories;

		/// <summary>
		/// Create an application item from a desktop file.
		/// </summary>
		/// <param name="desktopFile">
		/// A <see cref="System.String"/> containing the absolute path of
		/// a desktop (.desktop) file.
		/// </param>
		protected ApplicationItem (IWshShortcut item)
		{
			this.item = item;

			name = Path.GetFileNameWithoutExtension (item.FullName);
			description = item.Description;
			categories = Enumerable.Empty<string> ();

			if (item.IconLocation.StartsWith (","))
				icon = item.TargetPath + item.IconLocation;
			else
				icon = item.IconLocation;
		}

		public override string Name
		{
			get { return name; }
		}

		public override string Description
		{
			get { return description; }
		}

		public override string Icon
		{
			get { return icon; }
		}

		public IEnumerable<string> Categories
		{
			get { return categories; }
		}

		public bool NoDisplay
		{
			get
			{
				return false;
			}
		}

		public string Exec
		{
			get { return item.TargetPath; }
		}

		protected string Location
		{
			get { return item.TargetPath; }
		}

		public bool Hidden
		{
			get { return false; }
		}

		public bool IsUserCustomItem
		{
			get { return false; }
		}

		/// <summary>
		/// Executes the application.
		/// </summary>
		public void Run ()
		{
			Process.Start (item.TargetPath, item.Arguments);
		}

		public void LaunchWithFiles (IEnumerable<IFileItem> files)
		{
			foreach (IFileItem file in files)
				Process.Start (file.Path);
		}
	}
}